# #!usr/bin/env python3
# # -*- coding: UTF-8 -*-
# # !usr/bin/env python3
# # -*- coding: UTF-8 -*-
# """
# 数据增强流程：
#     读取候选背景图片列表  ——>  读取图片   ——> 读取标注信息 ——> 截取区域数据 ——> 确定缩放系数 ——>确定嵌入坐标 ——>保持图片和xml文件
# 可配置参数：
#     除嵌入外的其他增强方式
#     每个框合成的增强样本数量
#     是否允许多目标混合
#     是否允许多类别混合
#     最大/最小缩放系数
# 文件名秒格式：
#     aug + box_index + aug_index + filename
# """
# import os
# import re
#
# import numpy as np
# from random import choice, shuffle
#
# from PIL import Image
# from tqdm import tqdm
# from xl_tool.xl_io import file_scanning, read_json
# # from xl_tool.xl_io import l
#
# from ..config import IMAGE_FORMAT
# from ..annonation import get_boximgs,Text2XML
# from ..general import  linear_contrast, grey_world, \
#     affine_with_rotate_scale
#
# aug_dict = {"contrast": linear_contrast, "grey": grey_world, "affine": affine_with_rotate_scale}
#
#
# def get_background_images(path):
#     files = file_scanning(path, full_path=True, file_format=IMAGE_FORMAT)
#     return files
#
#
# def get_xml_image(image_path, xml_path):
#     valid_images = []
#     valid_xmls = []
#     image_files = file_scanning(image_path, file_format=IMAGE_FORMAT, full_path=True)
#     for image_file in image_files:
#         xml_file = xml_path + "/" + os.path.basename(image_file).split(".")[0] + ".xml"
#         if os.path.exists(xml_file):
#             valid_images.append(image_file)
#             valid_xmls.append(xml_file)
#     return valid_images, valid_xmls
#
#
# def get_propotions(cat):
#     big_propotion_cats = {"pizza"}
#     small_propotion_cats = {"chicken_wing", "drumstick", "toast"}
#     tiny_propotion_cats = {"egg_tart", }
#     if cat in big_propotion_cats:
#         return 0.8, 0.8
#     elif cat in small_propotion_cats:
#         return 0.6, 0.5
#     elif cat in tiny_propotion_cats:
#         return 0.4, 0.4
#     else:
#         return 0.7, 0.6
#
#
# def aug_images_from_xml(cat_name, image_files, xml_files, background_path, cat_aug_path, xml_folder,
#                         xml_source, x_proportion=0.4, y_proportion=0.5, boxes_per_image=1, max_num=1000,
#                         min_box_edge=100):
#     background_images = get_background_images(background_path)
#     im_xmls = list(zip(image_files, xml_files))
#     shuffle(im_xmls)
#     pbar = tqdm(im_xmls)
#     count = 0
#     for image_file, xml_file in pbar:
#         if count >= max_num:
#             continue
#         base_name = os.path.basename(image_file).split(".")[0]
#         boximgs = get_boximgs(image_file, xml_file)
#         boximgs = [boximg_info for boximg_info in boximgs if boximg_info["name"] == cat_name]
#         shuffle(boximgs)
#         box_count = 0
#         for i, boximg_info in enumerate(boximgs, 1):
#             if box_count > boxes_per_image:
#                 break
#             if sum(boximg_info["img"].size) < min_box_edge:
#                 # print("图片过小，不适合增强！",boximg_info["img"].size)
#                 continue
#             background_img_array = np.array(Image.open(choice(background_images)))
#             aug_image, coordinate = blending_one_image("", background_img_array=background_img_array,
#                                                        blending_img_array=np.array(boximg_info["img"]),
#                                                        x_proportion=x_proportion,
#                                                        y_proportion=y_proportion)
#             objects_info = [[boximg_info["name"]] + coordinate]
#             text2xml = Text2XML()
#             boximg_file = "aug_{}_{}_{}.jpg".format(i, 0, base_name)
#             xml = text2xml.get_xml(xml_folder, boximg_file, boximg_file, xml_source, aug_image.size, objects_info)
#             boxxml_file = "aug_{}_{}_{}.xml".format(i, 0, base_name)
#             aug_image.save(cat_aug_path + "/" + boximg_file)
#             with open(cat_aug_path + "/" + boxxml_file, "w") as f:
#                 f.write(xml)
#             count += 1
#             box_count += 1
#
#         pbar.set_description("图片增强进度：")
#
#
# def aug_images_from_image(blending_method, image_files, background_path, cat_aug_path, xml_folder, xml_source,
#                           cat_name, target_number=None, x_proportion=0.4, y_proportion=0.5, max_num=10000,
#                           simple_augs=("contrast", "grey", "affine", None), min_size_sum=200):
#     background_images = get_background_images(background_path)
#     # background_images = sorted(background_images, key=lambda x:int(os.path.basename(x).split(".")[0]))[18:]
#     if min_size_sum:
#         image_files = [file for file in image_files if sum(Image.open(file).size) > min_size_sum]
#     pbar = tqdm(image_files) if not target_number else tqdm([choice(image_files) for _ in range(target_number)])
#     count = 0
#     for image_file in pbar:
#         if count >= max_num:
#             continue
#         base_name = os.path.basename(image_file).split(".")[0]
#         base_name = cat_name.replace(" ", "_") if re.search(r"[^a-zA-Z0-9_\-.]", base_name) else base_name
#         bc = choice(background_images)
#         background_img_array = np.array(Image.open(bc))
#         if sum(background_img_array.shape[:2]) < 200:
#             continue
#         save_img = f"{cat_aug_path}/aug_{1}_{count}_{blending_method}_{os.path.basename(bc).split('.')[0]}_{base_name}.jpg"
#         aug_image, coordinate = blending_one_image(blending_method, background_img_array=background_img_array,
#                                                    blending_img_array=np.array(Image.open(image_file)),
#                                                    x_proportion=x_proportion,
#                                                    y_proportion=y_proportion, save_img=save_img)
#         if simple_augs:
#             aug_method = choice(simple_augs)
#             if aug_method:
#                 image_array = np.array(Image.open(save_img))
#                 Image.fromarray(aug_dict[aug_method](image_array)).save(save_img)
#         text2xml = Text2XML()
#         objects_info = [[cat_name] + coordinate]
#         boximg_file = os.path.basename(save_img)
#         xml = text2xml.get_xml(xml_folder, boximg_file, boximg_file, xml_source, aug_image.size, objects_info)
#         boxxml_file = boximg_file.replace("jpg", "xml")
#         aug_image.save(cat_aug_path + "/" + boximg_file)
#         with open(cat_aug_path + "/" + boxxml_file, "w") as f:
#             f.write(xml)
#         count += 1
#         pbar.set_description("图片增强进度：")
#
#
# def aug_images_mul_object(object_path, background_config_file, save_path, target_number=5000,
#                           distribute=(0.05, 0.35, 0.4, 0.15)):
#     """
#     流程：选择目标
#     """
#     folder = r"Food2019"
#     source = 'FoodDection'
#     cats = os.listdir(object_path)
#     cats.remove("empty")
#     background_configs = read_json(background_config_file)
#     # background_config = x
#     cats_files = {cat: file_scanning(f"{object_path}/{cat}", file_format=IMAGE_FORMAT) for cat in cats}
#     count = 0
#     cats_count = {cat: 0 for cat in cats}
#     while count < target_number:
#         ratio = count / target_number
#         if ratio <= distribute[0]:
#             sp_dis = 1
#             number = 1
#         elif distribute[0] < ratio <= distribute[1]:
#             sp_dis = choice([1, 2])
#             number = 2
#         elif distribute[1] < ratio <= distribute[2]:
#             sp_dis = choice([1, 2])
#             number = 3
#         else:
#             sp_dis = 1
#             number = 4
#         choose_cats = [choice(cats) for _ in range(number)]
#         for c in choose_cats:
#             cats_count[c] += 1
#         choose_files = [choice(cats_files[cat]) for cat in choose_cats]
#         background_config = choice(background_configs)
#         background_file, region = background_config["file"], background_config['region']
#         save_file = f"{save_path}/aug_{number}_{count}.jpg"
#         background_img, image_sizes, positions = blending_images(background_file, choose_files, sp_dis, save_file,
#                                                                  region)
#         objects_info = [[cat, ] + list(positions[i]) for i, cat in enumerate(choose_cats)]
#         text2xml = Text2XML()
#         boximg_file = f"aug_{number}_{count}.jpg"
#         xml = text2xml.get_xml(folder, boximg_file, boximg_file, source, background_img.size, objects_info)
#         boxxml_file = f"{save_path}/aug_{number}_{count}.xml"
#         with open(boxxml_file, "w") as f:
#             f.write(xml)
#         count += 1
#         # print(count)
#     print(cats_count)
#
#
# def test_single_xml():
#     data_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建" \
#                 r"数据集\3_公开数据集抽取\原始标注文件"
#     aug_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建" \
#                r"数据集\3_公开数据集抽取\增强文件"
#     background_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\data" \
#                       r"set\自建数据集\4_背景底图"
#     folder = r"Food2019"
#     source = 'FoodDection'
#     dirs = os.listdir(data_path)
#     for dir_ in dirs:
#         x_proportion, y_proportion = get_propotions(dir_)
#         cat_path = data_path + "/" + dir_
#         cat_aug_path = aug_path + "/" + dir_
#         os.makedirs(cat_aug_path, exist_ok=True)
#         image_files, xml_files = get_xml_image(cat_path, cat_path)
#         print("类别：{}\t有效标注文件数量：{}".format(dir_, len(image_files)))
#         aug_images_from_xml(dir_, image_files, xml_files, background_path, cat_aug_path, folder, source,
#                             x_proportion=x_proportion, y_proportion=y_proportion)
#
#
# def test_single():
#     data_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\5_抽取目标\网络"
#     aug_path_d = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\7_增强图片\单类别_直接融合"
#     aug_path_p = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\7_增强图片\单类别_金字塔融合"
#     background_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\data" \
#                       r"set\自建数据集\4_背景底图"
#     folder = r"Food2019"
#     source = 'FoodDection'
#     dirs = os.listdir(data_path)
#     for dir_ in dirs:
#         x_proportion, y_proportion = get_propotions(dir_)
#         cat_path = data_path + "/" + dir_
#         cat_aug_path_p = aug_path_p + "/" + dir_
#         cat_aug_path_d = aug_path_d + "/" + dir_
#         os.makedirs(cat_aug_path_p, exist_ok=True)
#         os.makedirs(cat_aug_path_d, exist_ok=True)
#         image_files = file_scanning(cat_path, file_format=IMAGE_FORMAT)
#         print("类别：{}\t有效标注文件数量：{}".format(dir_, len(image_files)))
#         # aug_images_from_image("", image_files, background_path, cat_aug_path_p, folder, source,cat_name=dir_,
#         #            x_proportion=x_proportion, y_proportion=y_proportion)
#         aug_images_from_image("direct", image_files, background_path, cat_aug_path_d, folder, source, cat_name=dir_,
#                               x_proportion=x_proportion, y_proportion=y_proportion)
#
#
# def test_mul_aug():
#     object_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\5_抽取目标"
#     background_config_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图\background_config.json"
#     save_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\7_增强图片\多类别组合"
#     aug_images_mul_object(object_path, background_config_file, save_path)
#
#
# def main():
#     # test_mul_aug()
#     test_single()
#
#
# if __name__ == '__main__':
#     main()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
    常用io读取函数
"""

import os
import re
import chardet
import random
# from aip import AipNlp
import time
import requests
import base64
import json
import numpy as np
from functools import wraps
from math import ceil
import logging
import zipfile


def ali_ocr_api(image_path, url="",
                appcode="",
                prob=False, charinfo=False, rotate=False,
                table=False, sortpage=False):
    """阿里巴巴图片识别API"""
    with open(image_path, "rb") as f:
        img = base64.b64encode(f.read()).decode()
    body = {
        "img": img,
        "prob": prob,  # 是否需要识别结果中每一行的置信度，默认不需要。 true：需要 false：不需要
        "charInfo": charinfo,  # 是否需要单字识别功能，默认不需要。 true：需要 false：不需要
        "rotate": rotate,  # 是否需要自动旋转功能，默认不需要。 true：需要 false：不需要
        "table": table,  # 是否需要表格识别功能，默认不需要。 true：需要 false：不需要
        "sortPage": sortpage  # 字块返回顺序，false表示从左往右，从上到下的顺序，true表示从上到下，从左往右的顺序，默认false
    }
    headers = {"Content-Type": "application/octet-stream; charset=utf-8", "Authorization": "APPCODE {}".format(appcode)}
    try:
        response = requests.post(url, json.dumps(body).encode(encoding='UTF8'), headers=headers)
        if response.status_code == 200:
            print("成功解析图片！！！")
            return response.json()
        else:
            print("访问错误：", response.status_code)
            return False
    except Exception as e:
        print("发生错误！！！")
        print(e)
        return False
def dict_transpose(dict_):
    """键值对对换"""
    return dict([j,i] for i,j in dict_.items())

def zip(path, zip_filename,compression=zipfile.ZIP_DEFLATED):
    """ compress file or directory"""
    z = zipfile.ZipFile(zip_filename, "w",compression)
    if os.path.isfile(path):
        z.write(path)
    else:
        for root, dirs, files in os.walk(path, topdown=True):
            if files:
                for file in files:
                    z.write(f"{root}/{file}")
            else:
                z.write(root)
    z.close()

def file_scanning(path, file_format=r".txt$", full_path=True, sub_scan=False):
    """
        scanning directory and return file paths with specified format
        :param path: directory to scan
        :param file_format:  file format to return ,regular patterns
        :param full_path: whether to return the full path
        :param sub_scan: whether to sanning the subfolder
        :return:file paths
        """
    if os.path.exists(path):
        file_paths = []
        for root, dirs, files in os.walk(path, topdown=True):
            paths = [file for file in files if re.search(file_format, file)]
            if full_path:
                paths = [os.path.join(root, file) for file in paths]
            file_paths.extend(paths)
            if not sub_scan:
                break
        if not file_paths:
            print("File with specified format not find")
            return []
    else:
        print("Invalid path!")
        return []
    return file_paths


def read_txt(file, return_list=False, remove_linebreak=True):
    """read txt, deal with encoding error(rt-utf-8-chardet)
    Args:
        file: full file path
        return_list: read by line and return a list
        remove_linebreak: only valid when return_list=True, remove line break for every line text
    Returns:
        text or list
        """
    if not os.path.exists(file):
        print("file not find!", file)
        return False
    try:
        with open(file, "rt") as f:
            if return_list:
                text = f.readlines()
                if remove_linebreak:
                    text = [t.rstrip("\n\r") for t in text]
            else:
                text = f.read()
    except (UnicodeDecodeError, UnicodeEncodeError):
        # print("encoding error, try encoding with utf-8")
        try:
            with open(file, "rt", encoding="utf-8") as f:
                if return_list:
                    text = f.readlines()
                    if remove_linebreak:
                        text = [t.rstrip("\n\r") for t in text]
                else:
                    text = f.read()
        except UnicodeDecodeError:
            # print("encoding error, try encoding with chardet, and ignore error")
            with open(file, "rb") as f:
                text_byte = f.read()
                encoding = chardet.detect(text_byte)["encoding"]
                text = text_byte.decode(encoding, errors="ignore")
                if return_list:
                    text = [i for i in text.split("\n")]
                    if remove_linebreak:
                        text = [t.rstrip("\n\r") for t in text]
        except PermissionError:
            print("not a valid path, please confirm")
            return False
    return text



# def baidu_nlp_client(APP_ID,API_KEY,SECRET_KEY):
    # """ 你的 APPID AK SK """
    # client = AipNlp(APP_ID, API_KEY, SECRET_KEY)
    # return client


# def ner_detect(text, net_type=("PER",), remove_blank=True):
    # """
    # Args:
    # :param text:检测文本
    # :param net_type: 实体检测类型，["ORG", "LOC", "PER"，"TIME"]
    # :param remove_blank:是否去除无用字符，空格代替
    # :return:
    # """
    # ner = {ner_name: [] for ner_name in net_type}  # set类型去重
    # if remove_blank:
        # text = re.sub(r"\W|\d", " ", text)  # 去掉数字、特殊字符、空字符等非汉字英文字符
    # if not text:
        # print("实体检测模块：未发现有效文本")
        # return ner
    # # print("发送实体检测文本：", text)
    # # time_s = datetime.now()
    # client = baidu_nlp_client()
    # try:
        # ner_result = client.lexer(text)
    # except requests.exceptions.ConnectionError:
        # time.sleep(0.2)
        # print("！！！百度实体检测接口连接失败尝试，尝试再次连接")
        # try:
            # ner_result = client.lexer(text)
        # except requests.exceptions.ConnectionError:
            # print("！！！百度实体检测接口连接失败尝试，返回空字典")
            # return ner
    # # print(ner_result)
    # try:
        # for item in ner_result["items"]:
            # if item["ne"] in net_type:
                # ner[item["ne"]].append(item["item"])
                # print("检测到实体: ", item["ne"], item["item"])
    # except KeyError:
        # print("未检测到有效文本")
    return ner


def time_analysis_wrapper(name="默认", show_time=True):
    """时间分析装饰器"""

    def time_analysis(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            st = time.time()
            result = func(*args, **kwargs)
            et = time.time()
            if show_time:
                print("———程序：{}———运行时间(s)：{:.4f}".format(name, (et - st)))
            return result

        return wrapper

    return time_analysis


@time_analysis_wrapper('实体识别')
def ner_detect_client(text, client, ner_type=("PER",), remove_blank=True):
    """百度实体识别封装接口
    Args:
        text:检测文本
        client: api客户端实例
        ner_type: 实体检测类型，["ORG", "LOC", "PER"，"TIME"]
        remove_blank: 是否去除非中文和英文字符
    Returns:
        字典形式{ner type:ner text}
    """
    ner2pos = {'PER': 'nr', 'LOC': 'ns', 'ORG': 'nt'}
    pos2ner = {'nr': 'PER', 'ns': 'LOC', 'nt': 'ORG'}
    ner = {ner_name: [] for ner_name in ner_type}
    ner_result = dict()
    pos_type = [ner2pos[i] for i in ner_type]
    if remove_blank:
        text = re.sub(r"\W|\d", " ", text)  # 去掉数字、特殊字符、空字符等非汉字英文字符
    if not text:
        print("实体检测模块：未发现有效文本")
        return ner
    for i in range(2):
        try:
            ner_result = client.lexer(text)
            break
        except requests.exceptions.ConnectionError:
            time.sleep(0.2)
            print("！！！百度实体检测接口连接失败尝试，尝试再次连接")
        # 特殊字符编码问题，直接替换掉特殊字符
        except UnicodeEncodeError:
            for ch in list(text):
                try:
                    ch.encode("gbk")
                except UnicodeEncodeError:
                    text = text.replace(ch, "")
            try:
                ner_result = client.lexer(text)
                break
            except Exception as e:
                print("出现未知错误:", e)
                return ner
    try:
        for item in ner_result["items"]:
            if item["ne"] in ner_type or item['pos'] in pos_type:
                ner_name = item["ne"] if item["ne"] else pos2ner[item["pos"]]
                ner[ner_name].append(item["item"])
                print("检测到实体: ", item["ne"], item["item"])
    except KeyError:
        print("未检测到有效文本")
    return ner




def to_categorical(y, num_classes=None):
    """Converts a class vector (integers) to binary class matrix.

    E.g. for use with categorical_crossentropy.

    # Arguments
        y: class vector to be converted into a matrix
            (integers from 0 to num_classes).
        num_classes: total number of classes.

    # Returns
        A binary matrix representation of the input. The classes axis
        is placed last.
    """
    y = np.array(y, dtype='int')
    input_shape = y.shape
    if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
        input_shape = tuple(input_shape[:-1])
    y = y.ravel()
    if not num_classes:
        num_classes = np.max(y) + 1
    n = y.shape[0]
    categorical = np.zeros((n, num_classes), dtype=np.float32)
    categorical[np.arange(n), y] = 1
    output_shape = input_shape + (num_classes,)
    categorical = np.reshape(categorical, output_shape)
    return categorical


def random_phone(mul_form=True):
    """随机生成11位有效的手机号，包含四种格式连接符
    Args:
        mul_form: 是否随机选择连接符如xxx-xxxx-xxxx的格式
    """
    full_num = '0123456789'
    # 手机号头三个数字有效性要求
    head_1 = ['13', '14', '15', '17', '18', '16', '19']
    head_2 = [full_num, '57', '012356789', '01235678', full_num, '2567', '189']
    index = random.randint(0, 6)
    head = head_1[index] + random.choice(head_2[index])
    middle = "".join([random.choice(full_num) for i in range(4)])
    tail = "".join([random.choice(full_num) for i in range(4)])
    # 连接格式
    random_form = ["-", "", " ", "—"]
    if mul_form:
        phone = random.choice(random_form).join([head, middle, tail])
    else:
        phone = head + middle + tail
    return phone


def random_email():
    """随机生成邮件地址"""
    dns = ["163.com", "126.com", "139.com", "qq.com", "sohu.com", "gmail.com", "foxmail.com", "outlook.com", "sina.com",
           "vip.163.com",
           "sina.cn", "qiye.aliyun.com", "21cn.com", "china.com", "bit.edu.cn", "ncu.edu.cn", "ustc.edu.cn",
           "corpease.net", "263.net", "shdongchang.com", "gmail.kenfor.net", "cumark.com.cn"]
    full = "0123456789qwertyuiopasdfghjklzxcvbnm"
    head_form = ["纯字母数字", "2段", "3段"]
    form = random.choice(head_form)
    if form == "纯数字字母":
        head = "".join([random.choice(full) for i in range(random.randint(6, 18))])
    elif form == "2段":
        head = "".join([random.choice(full[10:]) for i in range(random.randint(3, 8))])
        head = head + "_" + "".join([random.choice(full) for i in range(random.randint(3, 8))])
    else:
        head = "".join([random.choice(full) for i in range(random.randint(3, 8))])
        head = head + "_" + "".join([random.choice(full[10:]) for i in range(random.randint(3, 8))])
        head = head + "_" + "".join([random.choice(full) for i in range(random.randint(3, 8))])
    tail = random.choice(dns)
    email = head + "@" + tail
    return email


def read_json(filename):
    """读取json格式的字典与ID对应表
    Args:
        filename:文件名
    Returns:
        json转换python数据类型的数据，无数据或路径错误返回False"""
    try:
        if os.path.basename(filename) == filename:
            filename = os.getcwd() + '/' + filename
        with open(filename, mode="r", encoding="utf-8", buffering=1000000000) as f:
            data = json.load(f)
            if data:
                return data
            else:
                print("未读取到有效json数据！")
                return False
    except (FileExistsError, FileNotFoundError):
        print("未找到文件或者文件路径不合法！")
        return False


def read_with_base64(file):
    with open(file, 'rb') as f:
        file_base64 = base64.b64encode(f.read()).decode("utf-8")
    return file_base64


def save_to_json(data, filename, indent=None):
    """保存到json文件中，indent是否缩进
    Args:
        filename: 文件名，推荐绝对路径，如只有文件名则保存当前路径下
        data:保存数据
        indent:json格式缩进"""
    try:
        if os.path.basename(filename) == filename:
            filename = os.getcwd() + '/' + filename
        with open(filename, mode="w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=indent)

    except (FileExistsError, FileNotFoundError):
        print("未找到文件或者文件路径不合法！")





#!usr/bin/env python3
# -*- coding: UTF-8 -*-
from copy import deepcopy

from PIL import Image, ImageDraw
from PIL.Image import BILINEAR
import random
import cv2
import numpy as np
from abc import abstractmethod
from xl_tool.data.image.general import resize_with_pad
from xl_tool.xl_general import binary_search_nearest_abs, file_scanning
from .annonation import get_bndbox


# Todo 小目标组合多个排列合成
# Todo 单目标比例范围
# Todo 指定位置合成
# Todo 分割合成边缘模糊之：泊松/金字塔等


class BlendingBase:
    def __init__(self):
        pass

    @abstractmethod
    def blending_one_image(self):
        pass

    @classmethod
    def max_blending_size(cls, background_size, x, y, x_proportion, y_proportion):
        """计算最大允许嵌入图片大小"""
        if x:
            max_x = min(int(background_size[0] * x_proportion), background_size[0] - x)
        else:
            max_x = int(background_size[0] * x_proportion)
        if y:
            max_y = min(int(background_size[1] * y_proportion), background_size[1] - y)
        else:
            max_y = int(background_size[1] * y_proportion)
        return max_x, max_y

    @classmethod
    def calculate_rescale(cls, background_size, blending_size, x, y, x_proportion, y_proportion):
        """计算融合嵌入图片缩放系数
        Args:
            background_size:背景图片尺寸，宽 X 高
            blending_size:背景图片尺寸，宽 X 高
            x:嵌入位置,宽度方向，可为空
            y:嵌入位置，高度方向，可为空
            x_proportion:宽度方向最大占比
            y_proportion:高度方向最大占比
        """
        max_x, max_y = cls.max_blending_size(background_size, x, y, x_proportion, y_proportion)
        x_scale = 1 if max_x > blending_size[0] else max_x / blending_size[0]
        y_scale = 1 if max_y > blending_size[1] else max_y / blending_size[1]
        scale = min(x_scale, y_scale)
        if scale < 0.99:
            pass
            # print("嵌入图像{}过大，需进行缩放，缩放比例：{:.2f}".format(blending_size, scale))
        return scale

    @classmethod
    def save_image(cls, img: Image.Image, save):
        try:
            img.save(save)
        except Exception as e:
            print("增强图片存储失败！:".format(save))
            pass

    @classmethod
    def valid_size_confirm(cls, image_sizes, region_size, sp_dis, min_rescale):
        """
        该函数暂时无用
        Args:
            image_sizes:尺寸列表
            region_size: 嵌入区域大小
            sp_dis ： 组合方式
            min_rescale: 最大缩放区域
        """
        if sp_dis in (1, 5):
            return (sum([size[0] for size in image_sizes]) * min_rescale) < region_size[0]
        elif sp_dis(2, 6):
            return (sum([size[1] for size in image_sizes]) * min_rescale) < region_size[1]
        elif sp_dis in (3, 4):
            for i in range(2, -1, -1):
                one_combine = image_sizes[i]
                valid_x = (one_combine[0] + max([image_sizes[j][0] for j in range(3) if j != i])) < region_size[0]
                valid_y = (one_combine[1] < region_size[1] and sum([image_sizes[j][1] for j in range(3) if
                                                                    j != i])) < region_size[1]
                if valid_x and valid_y:
                    return True
            return False
        else:
            return False
        pass

    @classmethod
    def paralleling_caculate(cls, image_sizes, region_size, sp_dis):
        # print("输入尺寸：", image_sizes, "\t区域大小:", region_size)
        first, second = (0, 1) if sp_dis in (1, 5) else (1, 0)
        first_sum = sum([size[first] for size in image_sizes])
        if first_sum < region_size[first]:
            rescales = [1 if size[second] < region_size[second] else region_size[second] / size[second] for size in
                        image_sizes]
        else:
            # print("图片过大，进行缩放！")
            rescale = region_size[first] * 0.95 / first_sum
            rescales = [rescale if size[second] * rescale < region_size[second] else region_size[second] / size[second]
                        for size in
                        image_sizes]
        image_sizes = [[int(c * rescale) for c in size] for (rescale, size) in zip(rescales, image_sizes)]
        margin_first_max = region_size[first] - sum([size[first] for size in image_sizes])
        margin_seconds = [region_size[second] - size[second] for size in image_sizes]
        positions = []
        margin_first_remain = margin_first_max
        first_start = 0
        for size, margin_second in zip(image_sizes, margin_seconds):
            position = [None, None, None, None]
            margin = int(margin_first_remain * random.random())
            # print(margin)
            margin_first_remain = margin_first_remain - margin
            position_first = margin + first_start
            position_second = int(margin_second * random.random())
            first_start = position_first + size[first]
            position[first], position[second], position[first + 2], position[second + 2] = \
                position_first, position_second, first_start, position_second + size[second]
            positions.append(position)
        # print("缩放比例：", rescales, "\n缩放后大小：", image_sizes)
        return image_sizes, rescales, positions

    @classmethod
    def calculate_mul_blending_positions(cls, image_sizes, region_size, sp_dis):
        """计算嵌入位置
            1，2为并列排列
            3为混合排列，根据尺寸随机切分成上下或者左右2块，然后再做并列排列（暂未完成）
        """
        if sp_dis in (1, 5, 2, 6):
            return cls.paralleling_caculate(image_sizes, region_size, sp_dis)
        else:
            pass

    def rescale_blending(self, blending, background, x, y, x_proportion,
                         y_proportion, x_shift, y_shift, blending_region=None):
        """
        Args:
            blending:
            background:
            x:
            y:
            x_proportion:
            y_proportion:
            x_shift:
            y_shift:

        Returns:
            (x,y,blending),
                x: 合成图片插入横坐标（对应宽度，PIL的size[0], opencv的shape[1]），左上角
                y: 合成图片插入坐标（对应高度，PIL的size[1], opencv的shape[0]），左上角
                blending: resize后的图片
        """
        blending_size = (blending.shape[1], blending.shape[0],)
        background_size = (background.shape[1], background.shape[0])
        if not blending_region:
            blending_region = [0, 0, *background_size]
        region_size = (blending_region[2] - blending_region[0], blending_region[3] - blending_region[1])
        assert region_size[0] > 0, '插入区域宽度小于0'
        assert region_size[1] > 0, '插入区域高度小于0'
        scale_whole = self.calculate_rescale(background_size, blending_size, x, y, x_proportion, y_proportion)
        scale_region = self.calculate_rescale(region_size, blending_size, x, y, 0.99, 0.99)
        scale = min(scale_whole, scale_region)
        new_size_cv = (int(blending_size[0] * scale), int(blending_size[1] * scale))
        blending = blending if 0.99 < scale < 1.01 else cv2.resize(blending, dsize=new_size_cv)
        if not x:
            x = blending_region[0] + int(((region_size[0] - blending.shape[1]) // 2) * random.uniform(*x_shift))
        if not y:
            y = blending_region[1] + int(((region_size[1] - blending.shape[0]) // 2) * random.uniform(*y_shift))
        return x, y, blending

        # blending_size = (blending.shape[1], blending.shape[0],)
        # background_size = (background.shape[1], background.shape[0])
        # scale = self.calculate_rescale(background_size, blending_size, x, y, x_proportion, y_proportion)
        # new_size_cv = (int(blending_size[0] * scale), int(blending_size[1] * scale))
        # blending = blending if 0.99 < scale < 1.01 else cv2.resize(blending, dsize=new_size_cv)
        # if not x:
        #     x = int(((background_size[0] - blending.shape[1]) // 2) * random.uniform(*x_shift))
        # if not y:
        #     y = int(((background_size[1] - blending.shape[0]) // 2) * random.uniform(*y_shift))
        # return x, y, blending


class DirectBlending(BlendingBase):
    """
    直接对图片进行融合嵌入，主要包括单目标融合和多目标融合
    """

    def __init__(self):
        super(DirectBlending).__init__()

    def blending_one_image(self, background_img_file=None, blending_img_file=None,
                           background_img_array=None, blending_img_array=None, x=None, y=None,
                           x_proportion=0.7, y_proportion=0.7, x_shift=(0.5, 1.5), blending_region=None,
                           y_shift=(1, 1.9), save_img="", resample=BILINEAR,test=False):
        """单张图片融合嵌入
        下一版完善功能：
            最小比例占比限制
            长宽比扰动
        Args:
            background_img_file:背景图片
            blending_img_file:嵌入图片
            background_img_array:背景图片数组, background_img_file为空时传递该值可用
            blending_img_array:融合图片数组，blending_img_file为空时传递该值可用
            x: 嵌入位置x坐标（左上角），如为空，根据中心位置计算
            y:同上
            x_shift:针对x中心位置的摆动,取值范围为0-2
            y_shift:针对y中心位置的摆动，取值范围为0-2
            x_proportion: 嵌入图片x长度与背景图片x长度的最大占比
            y_proportion: 嵌入图片y长度与背景图片y长度的最大占比
            resample: 图像插值方式
            save_img:是否保持图片，是则输入路径，否则为空字符
        Returns:
            数组：Image格式的融合图片 + 融合后的bounding box坐标
                blending image,[xmin,ymin,xmax,ymax]
        """
        # background_img = Image.open(background_img_file) if background_img_array is None else \
        #     Image.fromarray(background_img_array)
        # blending_img = Image.open(blending_img_file) if blending_img_array is None else \
        #     Image.fromarray(blending_img_array)
        # if not blending_region:
        #     blending_region = [0,0,*blending_img.size]
        # bg_size = background_img.size
        # region_size = (blending_region[2]-blending_region[0],blending_region[3]-blending_region[1])
        # scale_whole = self.calculate_rescale(bg_size, blending_img.size, x, y, x_proportion, y_proportion)
        # scale_region = self.calculate_rescale(region_size, blending_img.size, x, y, x_proportion, y_proportion)
        # new_size = (int(size[0] * scale), int(size[1] * scale))
        # blending_img = blending_img if 0.99 < scale < 1.01 else blending_img.resize(new_size, resample=resample)
        # if not x:
        #     x = int(((background_img.size[0] - blending_img.size[0]) // 2) * random.uniform(*x_shift))
        # if not y:
        #     y = int(((background_img.size[1] - blending_img.size[1]) // 2) * random.uniform(*y_shift))

        background = cv2.imdecode(np.fromfile(background_img_file, dtype=np.uint8), cv2.IMREAD_UNCHANGED) \
            if background_img_array is None else background_img_array[:, :, ::-1]
        blending = cv2.imdecode(np.fromfile(blending_img_file, dtype=np.uint8), cv2.IMREAD_UNCHANGED) \
            if blending_img_array is None else blending_img_array[:, :, ::-1]
        # print(blending.shape)
        x, y, blending = self.rescale_blending(blending, background, x, y,
                                               x_proportion, y_proportion, x_shift, y_shift,blending_region)

        background_img = Image.fromarray(background[:, :, ::-1])
        blending_img = Image.fromarray(blending[:, :, ::-1])
        background_img.paste(blending_img, (x, y))
        blending_result = background_img
        if save_img:
            self.save_image(blending_result, save_img)
        if test:
            blending_result.show()
        return blending_result, [x, y, x + blending_img.size[0], y + blending_img.size[1]]

    def blending_images(self, background_img_file, blending_img_files, sp_dis, save_img="", blending_region=None):
        """
        多目标融合：
        多目标融合的位置分布主要包括以下几种方式：
            左右并列排列  1
            上下并列排列  2
            混合排列   3
        """
        background_img = Image.open(background_img_file)
        # print("背景图片大小：", background_img.size)
        blending_imgs = [Image.open(blending_img_file) for blending_img_file in blending_img_files]
        image_sizes = [img.size for img in blending_imgs]
        blending_region = blending_region if blending_region else (0, 0, background_img.size[0], background_img.size[1])
        x, y = blending_region[0], blending_region[1]
        region_size = blending_region[2] - blending_region[0], blending_region[3] - blending_region[1]
        image_sizes, rescales, positions = self.calculate_mul_blending_positions(image_sizes, region_size, sp_dis)
        positions = [(position[0] + x, position[1] + y, position[2] + x, position[3] + y) for position in positions]
        # print("嵌入坐标位置：", positions)
        blending_imgs = [blending_img.resize(size) for (blending_img, size) in zip(blending_imgs, image_sizes)]
        for blending_img, position in zip(blending_imgs, positions):
            # print(blending_img.size, position)
            background_img.paste(blending_img, (position[0], position[1]))
        if save_img:
            self.save_image(background_img, save_img)
        # background_img.show()
        return background_img, image_sizes, positions


class PyramidBlending(BlendingBase):
    """
    Args:
        num_pyramid: 层数，默认为6
        min_size； 金字塔最后一层最小尺寸，默认为6
    """

    def __init__(self, num_pyramid=6, min_size=4):
        super(PyramidBlending).__init__()
        self.num_pyramid = num_pyramid
        self.min_size = min_size
        pass

    def generate_gaussian_pyramid(self, image_array):
        """计算高斯金字塔"""
        array = image_array.copy()
        sizes = []
        if min(image_array.shape[:2]) ** (1 / (2 ** self.num_pyramid)) < (self.min_size - 1):
            base_exp = (min(image_array.shape[:2]) / self.min_size) ** (1 / self.num_pyramid)
            denominators = [base_exp ** i for i in range(1, self.num_pyramid + 1)]
            sizes = [(int(image_array.shape[0] / i), int(image_array.shape[1] / i)) for i in denominators]
        gaussian_pyramid = [array]
        for i in range(self.num_pyramid):
            array = cv2.pyrDown(array, sizes[i]) if sizes else cv2.pyrDown(array)
            gaussian_pyramid.append(array)
        return gaussian_pyramid

    def generate_laplacian_pyramid(self, gaussian_pyramid):
        """计算拉普拉斯金字塔"""
        laplacian_pyramid = [gaussian_pyramid[self.num_pyramid - 1]]
        for i in range(self.num_pyramid - 1, 0, -1):
            ge = cv2.pyrUp((gaussian_pyramid[i]),
                           dstsize=(gaussian_pyramid[i - 1].shape[1], gaussian_pyramid[i - 1].shape[0]))
            laplacian = cv2.subtract(gaussian_pyramid[i - 1], ge)
            laplacian_pyramid.append(laplacian)
        return laplacian_pyramid

    def surround_reconstruct(self, lp_background, lp_blending, normalize_x, normalize_y):
        ls = []
        for la, lb in zip(lp_background, lp_blending):
            temp = deepcopy(la)
            x, y = int(normalize_x * la.shape[1]), int(normalize_y * la.shape[0])
            temp[y:y + lb.shape[0], x:x + lb.shape[1]] = lb
            ls.append(temp)
        ls_array = ls[0]
        for i in range(1, self.num_pyramid):
            ls_array = cv2.pyrUp(ls_array, dstsize=(ls[i].shape[1], ls[i].shape[0]))
            ls_array = cv2.add(ls_array, ls[i])
        return ls_array

    def blending_one_image(self, background_img_file=None, blending_img_file=None,
                           background_img_array=None, blending_img_array=None,
                           x=None, y=None, x_proportion=0.6, y_proportion=0.6, blending_region=None,
                           x_shift=(0.5, 1.5), y_shift=(1.0, 1.9), save_img="",test=False):
        """基于opencv的金字塔融合"""
        background = cv2.imdecode(np.fromfile(background_img_file, dtype=np.uint8), cv2.IMREAD_UNCHANGED) \
            if background_img_array is None else background_img_array[:, :, ::-1]
        blending = cv2.imdecode(np.fromfile(blending_img_file, dtype=np.uint8), cv2.IMREAD_UNCHANGED) \
            if blending_img_array is None else blending_img_array[:, :, ::-1]
        # print(blending.shape)
        x, y, blending = self.rescale_blending(blending, background, x, y,
                                               x_proportion, y_proportion, x_shift, y_shift,blending_region=blending_region)

        g_background = self.generate_gaussian_pyramid(background)
        g_blending = self.generate_gaussian_pyramid(blending)
        lp_background = self.generate_laplacian_pyramid(g_background)
        lp_blending = self.generate_laplacian_pyramid(g_blending)
        blending_result = Image.fromarray(
            self.surround_reconstruct(lp_background, lp_blending,
                                      x / background.shape[1], y / background.shape[0])[:, :, ::-1])
        if save_img:
            self.save_image(blending_result, save_img)
        if test:
            blending_result.show()
        return blending_result, [x, y, x + blending.shape[1], y + blending.shape[0]]


class PoissonBlending(BlendingBase):

    def __init__(self):
        super(PoissonBlending).__init__()
        pass

    def blending_one_image(self, background_img_file=None, blending_img_file=None,
                           background_img_array=None, blending_img_array=None,
                           x=None, y=None, x_proportion=0.6, y_proportion=0.6, blending_region=None,
                           x_shift=(0.5, 1.5), y_shift=(1.0, 1.9), save_img="",test=False):
        """基于opencv的金字塔融合"""
        background = cv2.imdecode(np.fromfile(background_img_file, dtype=np.uint8),
                                  cv2.IMREAD_UNCHANGED) \
            if background_img_array is None else background_img_array[:, :, ::-1]
        blending = cv2.imdecode(np.fromfile(blending_img_file, dtype=np.uint8), cv2.IMREAD_UNCHANGED) \
            if blending_img_array is None else blending_img_array[:, :, ::-1]
        # print(blending.shape)
        x, y, blending = self.rescale_blending(blending, background, x, y,
                                               x_proportion, y_proportion, x_shift, y_shift,blending_region=blending_region)
        # print(blending.shape)
        blending_result = Image.fromarray(
            cv2.seamlessClone(blending, background, np.ones(blending.shape, dtype=np.uint8) * 255,
                              (x + blending.shape[1] // 2, y + blending.shape[0] // 2),
                              flags=cv2.NORMAL_CLONE)[:, :, ::-1])
        if save_img:
            self.save_image(blending_result, save_img)
        if test:
            blending_result.show()
        return blending_result, [x, y, x + blending.shape[1], y + blending.shape[0]]


class ObjectReplaceBlend:
    """目标框替换"""

    def __init__(self):
        pass

    def blending_one_image(self, image_file, object_images, object_aspects, xml_file, classes=None,
                           random_choice=False, aspect_jump=0):
        """ """
        boxes = get_bndbox(xml_file)
        if classes:
            boxes = [i for i in boxes if i["name"] if classes]
        image = Image.open(image_file)
        aug_boxes = []
        for box in boxes:
            center_x = (box["coordinates"][2] + box["coordinates"][0]) / 2
            center_y = (box["coordinates"][3] + box["coordinates"][1]) / 2
            width = box["coordinates"][2] - box["coordinates"][0]
            height = box["coordinates"][3] - box["coordinates"][1]
            aspect = (width) / (height)
            if aspect_jump:
                aspect = aspect + random.uniform(-aspect_jump * 0.8, aspect_jump)
            best_aspect_index = binary_search_nearest_abs(object_aspects, aspect, 0, len(object_aspects) - 1)
            if random_choice:
                paste_image = random.choice(object_images)
            else:
                paste_image = object_images[best_aspect_index]
            paste_image, size = resize_with_pad(paste_image, target_size=(
                box["coordinates"][2] - box["coordinates"][0], box["coordinates"][3] - box["coordinates"][1]))
            image.paste(paste_image, box["coordinates"])
            aug_boxes.append([int(center_x - size[0] / 2),
                              int(center_y - size[1] / 2),
                              int(center_x + size[0] / 2),
                              int(center_y + size[1] / 2)])
        return image, aug_boxes


class SegBlend(BlendingBase):
    """
    分割图片合成
    """

    def __init__(self):
        super(SegBlend).__init__()
        pass

    def blending_one_image(self, background_img_file=None, blending_img_file=None, mask_img_file=None,
                           background_img_array=None, blending_img_array=None, mask_img_array=None,
                           x=None, y=None, x_proportion=0.6, y_proportion=0.6, poisson=False, blending_region=None,
                           x_shift=(0.5, 1.5), y_shift=(1.0, 1.9), save_img="", resample=BILINEAR):
        background_img = Image.open(background_img_file) if background_img_array is None else \
            Image.fromarray(background_img_array)
        blending_img = Image.open(blending_img_file) if blending_img_array is None else \
            Image.fromarray(blending_img_array)
        mask_img = Image.open(mask_img_file) if mask_img_array is None else \
            Image.fromarray(mask_img_array)
        size = blending_img.size
        scale = self.calculate_rescale(background_img.size, blending_img.size, x, y, x_proportion, y_proportion)
        new_size = (int(size[0] * scale), int(size[1] * scale))
        blending_img = blending_img if 0.99 < scale < 1.01 else blending_img.resize(new_size, resample=resample)
        mask_img = mask_img if 0.99 < scale < 1.01 else mask_img.resize(new_size, resample=resample)
        if not x:
            x = int(((background_img.size[0] - blending_img.size[0]) // 2) * random.uniform(*x_shift))
        if not y:
            y = int(((background_img.size[1] - blending_img.size[1]) // 2) * random.uniform(*y_shift))
        if not poisson:
            background_img.paste(blending_img, (x, y), mask_img)
            blending_result = background_img
        else:
            blending = np.array(blending_img)[:, :, ::-1]
            background = np.array(background_img)[:, :, ::-1]
            mask = np.array(mask_img)[:, :, ::-1] if mask_img.mode not in ["P", "L", "1"] else np.array(mask_img)
            mask = mask.astype(np.uint8) * 255
            blending_result = Image.fromarray(
                cv2.seamlessClone(blending, background, mask,
                                  (x + blending.shape[1] // 2, y + blending.shape[0] // 2),
                                  flags=cv2.NORMAL_CLONE)[:, :, ::-1])
        if save_img:
            self.save_image(blending_result, save_img)
        blending_result.show()
        return blending_result, [x, y, x + blending_img.size[0], y + blending_img.size[1]]

    def segbox_extract(self):
        """open image segmatation area extract"""
        pass


def test_blending_pyramid():
    background_img_file = r"D:\1.jpg"
    blending_img_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\public_dataset\food101\food\pizza" \
                        r"\702165.jpg"
    blender = PyramidBlending(6)
    blender.blending_one_image(background_img_file, blending_img_file)


def test_blending_direct():
    background_img_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图" \
                        r"\4.jpg"
    blending_img_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图" \
                        r"\55555.jpg"
    blender = DirectBlending()
    blender1 = PoissonBlending()
    blender2 = PyramidBlending()
    # for i in range(3):
    blender.blending_one_image(background_img_file, blending_img_file,blending_region=[100,200,540,480],test=True)
    blender1.blending_one_image(background_img_file, blending_img_file, blending_region=[100, 200, 540, 480],test=True)
    blender2.blending_one_image(background_img_file, blending_img_file, blending_region=[100, 200, 540, 480],test=True)

def test_mul_blending():
    a = DirectBlending()
    backgroud_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图\2.jpg"
    blending_files = [
        r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\5_抽取目标\broccoli\000000000009.jpg",
        r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\5_抽取目标\broccoli\000000000009.jpg",
        r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\5_抽取目标\orange\000000000009.jpg"]
    print(a.blending_images(backgroud_file, blending_files, 2, (20, 80, 620, 440)))


def test_mask_speed():
    import numpy as np
    import time
    b = np.random.randint(0, 256, (224, 224, 3), dtype=np.uint8)
    m = np.random.randint(0, 2, (224, 224, 3), dtype=np.uint8)
    b_img = Image.fromarray(b)
    m_img = Image.fromarray(m[:, :, 0], mode="1")
    st = time.time()
    for i in range(1000):
        b_img.paste(m_img, mask=m_img)
    print("paste时间：", time.time() - st)
    st = time.time()
    for i in range(1000):
        t = b * m
    print("numpy点乘时间：", time.time() - st)
    st = time.time()
    for i in range(1000):
        np.putmask(b, m, m)
        # t = b*m
    print("numpy掩码：", time.time() - st)


def test_seg_mask():
    blender = SegBlend()
    background_img_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图\1.jpg"
    blending_img_file = r"F:\Large_dataset\segmentation\mask\crop_00c1d2697deeaea1_m014j1m.jpg"
    mask_img_file = r"F:\Large_dataset\segmentation\image\crop_00bb5720a7ba062e_m014j1m_b26b27f6.png"
    blender.blending_one_image(background_img_file, blending_img_file, mask_img_file, poisson=True, x_proportion=0.2,
                               save_img=r"F:\Large_dataset\segmentation\segmentation_blend\crop_00bb5720a7ba062e_m014j1m_b26b27f6.jpg")


def test_ObjectReplaceBlend():
    for i in range(10):
        blender = ObjectReplaceBlend()
        base = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\1_真实场景\0_已标框\bacon"
        image_file = f"{base}/0a634ce3-4f09-5bd4-99a0-238331e88fad.jpg"
        xml_file = f"{base}/0a634ce3-4f09-5bd4-99a0-238331e88fad.xml"
        object_path = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\5_抽取目标\混合\bacon"
        object_images = [Image.open(i) for i in file_scanning(object_path, "jpg")]
        aspects = [i.size[0] / i.size[1] for i in object_images]
        object_images, aspects = zip(*sorted(zip(object_images, aspects), key=lambda x: x[1]))
        im, boxes = blender.blending_one_image(image_file, object_images, aspects, xml_file, random_choice=False,
                                               aspect_jump=1.0)
        draw = ImageDraw.Draw(im)
        for box in boxes:
            draw.rectangle(box)
        im.show()


def main():
    # blender = ObjectReplaceBlend()

    # blender = PoissonBlending()

    # background_img_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图\1.jpg"
    # blending_img_file = r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集" \
    #                     r"\5_抽取目标\混合\sweet_potato\baidu_sweet_potato_1027493905_3011920384.jpg"
    # a = blender.blending_one_image(background_img_file, blending_img_file, x=100, y=100)[0]
    # a.show()
    # print(a)
    # test_blending_pyramid()
    # test_blending_direct()
    # test_mul_blending()
    # test_seg_mask()
    test_ObjectReplaceBlend()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

"""
    多线程模块api
"""

import os
from functools import wraps
import threading
from math import ceil
import logging


class MyThread(threading.Thread):
    """带返回值的多线程类"""
    def __init__(self, target, args=()):
        super(MyThread, self).__init__()
        self.target = target
        self.args = args
        self.result = ""

    # 重新定义带返回值的线程类
    def run(self):
        self.result = self.target(*self.args)

    def get_result(self):
        try:
            return self.result
        except Exception as e:
            logging.warning(str(e))
            return None


def mul_thread(function, thread_param, share_param=(), thread_num=8):
    threads = list(range(thread_num))
    handle_num = ceil(len(thread_param) / thread_num)
    for i in range(thread_num):
        args = [thread_param[i * handle_num:(i + 1) * handle_num]]
        args.extend(list(share_param))
        threads[i] = MyThread(target=function,
                              args=tuple(args))
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="xl_tool", # Replace with your own username
    version="0.0.9",
    author="Xiaolin",
    author_email="119xiaolin@163.com",
    description="my tool function",
    long_description=long_description,
    long_description_content_type="text/markdown",
	install_requires=['requests>=2.21.0','numpy>=1.16.0','chardet>=3.0.4'],
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
	
)


#!usr/bin/env python3
# -*- coding: UTF-8 -*-
import os
import shutil

from PIL import Image
from cv2 import namedWindow, imshow, waitKey, WINDOW_FREERATIO, destroyAllWindows
from tqdm import tqdm
from skimage.metrics import structural_similarity as compare_ssim
import numpy as np
from random import uniform
import imgaug.augmenters as iaa
from xl_tool.xl_io import time_analysis_wrapper


def read_with_rgb(image_file):
    """将非rgb格式图片转成rgb"""
    img = Image.open(image_file)
    if img.mode != "RGB":
        img = img.convert("RGB")
    return img


def grey_world(img_array):
    """白平衡处理函数
    Args:
        img_array:输入三通道图像数组
    """
    R = img_array[:, :, 0].mean()
    G = img_array[:, :, 1].mean()
    B = img_array[:, :, 2].mean()
    avg = (B + G + R) / 3
    img_array[:, :, 0] = np.minimum(img_array[:, :, 0] * (avg / R), 255)
    img_array[:, :, 1] = np.minimum(img_array[:, :, 1] * (avg / G), 255)
    img_array[:, :, 2] = np.minimum(img_array[:, :, 2] * (avg / B), 255)
    return img_array.astype(np.uint8)


def linear_contrast(img_array, offset=0.3):
    """对比度：线性变换"""
    value = offset * uniform(-1, 1) + 1
    try:
        seq = iaa.Sequential([
            iaa.LinearContrast(value)
        ])
    except AttributeError:
        np.random.bit_generator = np.random._bit_generator
        seq = iaa.Sequential([
            iaa.LinearContrast(value)
        ])
    return seq(images=[img_array])[0]


def affine_with_rotate_scale(img_array, x_scale=(0.9, 1.1), y_scale=(0.9, 1.1), rotate=(-5, 5)):
    """简单仿射变换，即旋转和缩放"""
    try:
        seq = iaa.Sequential([iaa.Affine(
            scale={"x": x_scale, "y": y_scale},  # scale images to 80-120% of their size, individually per axis
            rotate=rotate,  # rotate by -45 to +45 degrees
            order=[0, 1]
        )])
    except AttributeError:
        np.random.bit_generator = np.random._bit_generator
        seq = iaa.Sequential([iaa.Affine(
            scale={"x": x_scale, "y": y_scale},  # scale images to 80-120% of their size, individually per axis
            rotate=rotate,  # rotate by -45 to +45 degrees
            order=[0, 1]
        )])

    return seq(images=[img_array])[0]


def resize_with_pad(image, target_size):
    """
    resize  with padding
    Args:
        image: 输入图片可以是PIL.Image也可以是numpy.ndarray
        target_size: 目标尺寸

    Returns:
        resized numpy or image( same type as input), scaled size with no padding
    """
    input_type = type(image)
    if type(image) == np.ndarray:
        image = Image.fromarray(image)
    if image.size == target_size:
        return (np.array(image), image.size) if input_type == np.ndarray else (image, image.size)
    else:
        scale = min(target_size[0] / image.size[0], target_size[1] / image.size[1])
        background = Image.new("RGB", target_size)
        image = image.resize([int(scale * i) for i in image.size])
        courdinate = (target_size[0] - image.size[0]) // 2, (target_size[1] - image.size[1]) // 2
        background.paste(image, courdinate)
        return (np.array(background), image.size) if input_type == np.ndarray else (background, image.size)


def cv_show_image(image_file, window_name="图片"):
    namedWindow(window_name, WINDOW_FREERATIO)  # 0表示压缩图片，图片过大无法显示
    imshow(window_name, image_file)
    k = waitKey(0)  # 无限期等待输入，需要有这个否则会死机
    if k == 27:  # 如果输入ESC退出
        destroyAllWindows()


#
# def blending_one_image(blending_method="direct", background_img_file=None, blending_img_file=None,
#                        background_img_array=None, blending_img_array=None,
#                        x=None, y=None, x_proportion=0.6, y_proportion=0.6,
#                        x_shift=(0.1, 1.9), y_shift=(0.5, 1.9), save_img=""):
#     blender = DirectBlending() if blending_method == "direct" else PyramidBlending()
#     blending_result, [x, y, x1, y1] = blender.blending_one_image(background_img_file, blending_img_file,
#                                                                  background_img_array, blending_img_array,
#                                                                  x, y, x_proportion, y_proportion, x_shift, y_shift,
#                                                                  save_img)
#     return blending_result, [x, y, x1, y1]
#
#
# def blending_images(background_img_file, blending_img_files, sp_dis, save_img="", blending_region=None,
#                     blending_method="direct"):
#     blender = DirectBlending() if blending_method == "direct" else PyramidBlending()
#     background_img, image_sizes, positions = blender.blending_images(background_img_file, blending_img_files, sp_dis,
#                                                                      save_img,
#                                                                      blending_region)
#     return background_img, image_sizes, positions


class ImageSimilarity:
    """
        Image similarity algorithm
            ssim:
            histogram similarity
            hash: hash similaritied,
            Todo : hash is note correct
    """

    def __init__(self):
        pass

    def hist_similar(self, image1, image2):
        """直方图计算"""
        lh, rh = image1.histogram(), image2.histogram()
        assert len(lh) == len(rh)
        return sum(1 - (0 if l == r else float(abs(l - r)) / max(l, r)) for l, r in zip(lh, rh)) / len(lh)

    def ssim(self, image1, image2):
        """ssim相似度计算"""
        return compare_ssim(np.array(image1), np.array(image2), multichannel=True)

    def dhash_sim(self, image1, image2):
        """计算差异哈希值算法的相似度"""
        # Todo:差异哈希未经仔细验证，有问题
        image1 = image1.resize((8, 8), Image.ANTIALIAS).convert('L')
        image2 = image2.resize((8, 8), Image.ANTIALIAS).convert('L')
        return self.hamming_distance(self.dHash(np.array(image1)), self.dHash(np.array(image2)))

    # 差异哈希算法
    def dHash(self, array):
        assert array.shape == (8, 8)
        hash = []
        # 每行前一个像素大于后一个像素为1，相反为0，生成哈希
        for i in range(8):
            for j in range(7):
                if array[i, j] > array[i, j + 1]:
                    hash.append(1)
                else:
                    hash.append(0)
        return hash

    # 均值哈希算法
    def aHash(self, array):
        assert array.shape == (8, 8)
        avreage = np.mean(array)
        hash = []
        for i in range(array.shape[0]):
            for j in range(array.shape[1]):
                if array[i, j] > avreage:
                    hash.append(1)
                else:
                    hash.append(0)
        return hash

    # 计算汉明距离
    def hamming_distance(self, hash1, hash2):
        num = 0
        for index in range(len(hash1)):
            if hash1[index] != hash2[index]:
                num += 1
        return num / len(hash1)

    @time_analysis_wrapper("相似度计算")
    def caculate_similarities(self, files, save=None, hash=False, hstm=False, resize=(48, 48)):
        images = [(os.path.basename(file), Image.open(file).resize(resize), file) for file in files]
        results = set()
        for i in tqdm(list(range(len(images) - 1))):
            for j in (list(range(i + 1, len(images)))):
                img1 = images[i][1]
                img2 = images[j][1]
                hstm = self.hist_similar(img1, img2) if hstm else 0
                ssim = self.ssim(img1, img2)
                dhash = self.dhash_sim(img1, img2) if hash else 0
                if ssim > 0.6 or hstm > 0.7:
                    print(f"\nsslim: {ssim}\thstm: {hstm}\tdhash:{dhash}\n-----",
                          images[i][0], "\t", images[j][0])
                    results.add(images[i][2])
                    results.add(images[j][2])
        if save:
            os.makedirs(save, exist_ok=True)
            for file in results:
                shutil.copy(file, save + "/" + os.path.basename(file))


def main():
    # embedding_img = Image.open(
    #     r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\3_公开数据集抽取\原始标注文件\hamburger\n07697313_12.JPEG")
    # background_img = Image.open(r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图\1.jpg")
    # aug = blending_one_image(background_img, embedding_img)
    # aug.show()
    image = Image.open(r"E:\Programming\Python\8_Ganlanz\food_recognition\dataset\自建数据集\4_背景底图\1.jpg")
    resize_with_pad(image, (256, 256)).show()
    # print(embedding_rescale((1000, 1000), (900, 900), x=None, y=None, x_proportion=0.7, y_proportion=0.7))


if __name__ == '__main__':
    main()
